module Geo.WayFinder 
    ( distance
    , fast_distance
    , dijkstra
    , find_path
    )
where

import List (sort)

import qualified World.Type as W
import Geo.SimpleFinder
import Geo.Gradient
import Geo.Util
import qualified Geo.Type (distance)
import Geo.Type hiding (distance)

import Util.Args 


-- distance by path calculation
distance :: Args -> W.Position -> [W.Position] -> IO([Int])
distance _ _ [] = do return []
distance args start (x:xs) = do
                             path  <- find_path args start x
			     prick <- Geo.WayFinder.distance args start xs
			     return $ (length path) : prick


-- calculates the distance via the nearest homebase with dijkstra
fast_distance :: Args -> W.Position -> [W.Position] -> IO([Int])
fast_distance _ _ [] = do return []
fast_distance args@Args { info=i
		   , distancemap=dmap
		   , waygraph=wgraph } start (x:xs) = do
						     ms_start <- get_membership i dmap start; -- start/end segment
						     let seg_start   = base_nr ms_start
						     let dist_start  = Geo.Type.distance ms_start
						     ms_target <- get_membership i dmap x
						     let seg_target  = base_nr ms_target
						     let dist_target = Geo.Type.distance ms_target
						     fm <- dijkstra args seg_target
						     -- use finitemap to cache the stuff
						     let path = lookupWithDefaultFM fm (0,[]) seg_target
						     let len_highway = sum $ map (\(_,way) -> Geo.Type.len way) $ 
								       snd $ path
						     prick <- fast_distance args start xs
--						     putStrLn $ show "path: " ++ show path 
--							      ++ show "len: " ++ show len
--							      ++ show "d_start: " ++ show dist_start
--							      ++ show "d_target: " ++ show dist_target
						     let len_path = dist_start + len_highway + dist_target-- case seg_start==seg_target of 
--							       True -> 
						     return $ len_path : prick


-- searching for way from one position to another one on the field (only significant files)
-- 4 possibilities:
--
--   start  - a intersection point, home base or another significant point
--          - another point
--
--   target - in a neighbor segment 
--          - in another segment
find_path :: Args -> W.Position -> W.Position -> IO([W.Position])
find_path args@Args { mappe=gmap
	       , info=i
	       , distancemap=dmap
	       , waygraph=wgraph } start target 
	  | (start == target) = do return []
	  | otherwise = do
			ms <- get_membership i dmap start -- start/end segment
			let seg_start  = base_nr ms
			ms <- get_membership i dmap target
			let seg_target = base_nr ms
			let dgraph = init_graph i wgraph seg_start
			let neighbor_edges = get_wgraph_edges_from wgraph seg_start
			case seg_start == seg_target of -- points are in the same segment
		 	 True  -> do 
				  path <- gradient i dmap start target
				  return $ tail path
			 False -> 
			     case null neighbor_edges of
			     -- target in neighbor segment
			      False -> do
				       path <- find_path_by_flood (gmap,i) start target
				       return $ tail path
				     -- target in another segment
			      True  -> do
				       p1   <- gradient i dmap start (get_homebase wgraph seg_start)
				       fm   <- dijkstra args seg_target
				       -- use finitemap to cache the stuff
				       let dijk = snd $ lookupWithDefaultFM fm (0,[]) seg_target
				       p_i  <- (mapM (\ (edge,way) -> do
				                                       locpath <- find_path_centre i dmap edge way
				                                       print locpath
						                       return $ tail locpath
			   			     ) dijk) :: IO([[W.Position]]) 
				       p3   <- gradient i dmap (get_homebase wgraph seg_target) target
				       return $ tail p1 ++ (concat p_i) ++ p3


-- initializes a graph for dijkstra with 0 and infinity (max_distance)
init_graph :: W.Info -> WayGraph Int -> Int -> Graph Int
init_graph i graph seg_start = Graph { gedges = edges graph
				     , gnodes = nodes graph
				     , ginfos = listToFM $ 
				       ((seg_start,(0,[]))
					: [ (homebase, (max_dist,[])) 
					    | homebase<-setToList (nodes graph)
					  , homebase/=seg_start
					  ]
				       )
				     } :: Graph Int
    where  max_dist = max_distance i


-- dijkstra to others visible
-- node -> distance,waylist
dijkstra :: Args -> Int -> IO(FiniteMap Int (Int,[(Edge W.Position, Way)]))
dijkstra Args {info=i, waygraph=wgraph} start = do 
					let dgraph = init_graph i wgraph start
					dijk <- dijkstra' wgraph start dgraph
					return dijk


-- internal implementation of dijkstra
--
-- dijkstra algorithm with usage of shortest path from WayGraph
-- returns the shortest path between homebases    
dijkstra' :: WayGraph Int -> Int -> Graph Int -> IO(FiniteMap Int (Int,[(Edge W.Position, Way)]))
-- end of recursion, so return the path of the target node
dijkstra' wgraph target g
	 | (isEmptySet (gnodes g)) = do 
				     return $ ginfos g
-- DEBUG
--				      error $ show "size: "++ show (sizeFM (ginfos g))
--					    ++ show "- keys: " ++ show (keysFM (ginfos g))
--					    ++ show "- target: "++ show target
--					    ++ show "- length: " ++ show t 
--					    ++ show "- path:  " ++ show p ++ "<-"
	 | otherwise = 
		dijkstra' wgraph target newgraph 
-- DEBUG
--	     error $ show (gedges newgraph) 
--		   ++ show (gnodes newgraph) 
--		   ++ show " - start:" ++ show x
--			  ++ show " - min:" ++ show minel
		where   -- 
			(x:xs) = minel : (filter (\n -> n/=minel) $ nlist)
			minel  = snd $ minimum $ map (\a -> (lookupWithDefaultFM (ginfos g) (0,[]) a
							    , a)
						     ) nlist
			nlist  = setToList $ gnodes g
			(start_dist,start_path) = lookupWithDefaultFM (ginfos g) (0,[]) x
			newgraph = Graph -- construct a new graph without the startnode
			   { gnodes     = mkSet xs
			   , gedges     = mkSet $ filter (\e -> from e/=x ) $ setToList $ gedges g 
			   , ginfos     = infos
			   } :: Graph Int
			infos = mapFM (\node (old_dist,old_path) -> 
							  case elementOf node neighbor_nodes of
								False -> (old_dist,old_path)
								True  -> 
								   let e   = Edge { from = get_homebase wgraph x 
										  , to   = get_homebase wgraph node
										  }
								       way = shortest_way wgraph (Edge x node)
								   in case start_dist + len way < old_dist of
									   True  -> ( start_dist + (len way) 
										    , start_path ++ [(e,way)]
										    )
									   False -> ( old_dist, old_path )
				      ) $ ginfos g
			neighbor_nodes = mkSet 
				   $ map (\e -> to e)
				   $ filter (\e -> from e==x) $ setToList (gedges g)

